import pytest
from rest_framework import serializers

from drf_jsonschema_serializer import JSONSchemaField, SerializerJSONField

schema = {
    "type": "object",
    "properties": {"a": {"type": "string"}, "b": {"type": "number"}},
}


format_schema = {
    "type": "object",
    "properties": {"a": {"type": "string", "format": "email"}},
}


def test_valid_json_field():
    field = JSONSchemaField(schema=schema)
    assert field.to_internal_value({"a": "foo", "b": 3.1}) == {"a": "foo", "b": 3.1}


def test_invalid_json_field():
    field = JSONSchemaField(schema=schema)
    with pytest.raises(serializers.ValidationError) as e:
        field.run_validation({"a": "foo", "b": "nonumber"})
    assert e.value.detail == ["'nonumber' is not of type 'number'"]


def test_json_field_valid_format_check():
    field = JSONSchemaField(schema=format_schema)
    assert field.to_internal_value({"a": "foo@example.com"}) == {"a": "foo@example.com"}


def test_json_field_invalid_format_check():
    field = JSONSchemaField(schema=format_schema)
    with pytest.raises(serializers.ValidationError) as e:
        field.run_validation({"a": "noemail"})
    assert e.value.detail == ["'noemail' is not a 'email'"]


def test_json_field_invalid_format_check_but_format_checking_disabled():
    field = JSONSchemaField(schema=format_schema, format_checker=None)
    assert field.to_internal_value({"a": "noemail"}) == {"a": "noemail"}


def test_valid_json_field_in_serializer():
    class TestSerializer(serializers.Serializer):
        json = JSONSchemaField(schema=schema)

    serializer = TestSerializer(data={"json": {"a": "foo", "b": 3.1}})
    assert serializer.is_valid()
    assert serializer.validated_data == {"json": {"a": "foo", "b": 3.1}}


def test_serialize_json_field_in_serializer():
    class TestSerializer(serializers.Serializer):
        json = JSONSchemaField(schema=schema)

    serializer = TestSerializer(data={"json": {"a": "foo", "b": 3.1}})
    assert serializer.is_valid()
    assert serializer.data == {"json": {"a": "foo", "b": 3.1}}


def test_invalid_json_field_in_serializer():
    class TestSerializer(serializers.Serializer):
        json = JSONSchemaField(schema=schema)

    serializer = TestSerializer(data={"json": {"a": "foo", "b": "nonumber"}})
    assert not serializer.is_valid()
    assert serializer.validated_data == {}


def test_serializer_json_field():
    class MySerializer(serializers.Serializer):
        foo = serializers.IntegerField()

    class TestSerializer(serializers.Serializer):
        json = SerializerJSONField(MySerializer)

    serializer = TestSerializer(data={"json": {"foo": 42}})
    assert serializer.is_valid()
    assert dict(serializer.validated_data) == {"json": {"foo": 42}}

    serializer = TestSerializer(data={"json": {"foo": "Not a number!"}})
    assert not serializer.is_valid()

    serializer = TestSerializer(data={"json": {"foo": "44"}})
    assert serializer.is_valid()
    assert dict(serializer.validated_data) == {"json": {"foo": 44}}


def test_serializer_json_field_error():
    class MySerializer(serializers.Serializer):
        foo = serializers.IntegerField()

    field = SerializerJSONField(MySerializer)
    with pytest.raises(serializers.ValidationError) as e:
        field.run_validation({"foo": "Not a number!"})
    # FIXME: is this really the right structure for errors in this
    # case?
    assert e.value.detail == {"foo": ["A valid integer is required."]}


def test_serializer_json_field_decimal():
    class MySerializer(serializers.Serializer):
        foo = serializers.DecimalField(None, 2)

    class TestSerializer(serializers.Serializer):
        json = SerializerJSONField(MySerializer)

    serializer = TestSerializer(data={"json": {"foo": "42.1"}})
    assert serializer.is_valid()
    assert dict(serializer.validated_data) == {"json": {"foo": "42.10"}}

    serializer = TestSerializer(data={"json": {"foo": "Not a number!"}})
    assert not serializer.is_valid()



# convert a serializer to a JSON Schema.
from rest_framework import serializers
from rest_framework.utils.field_mapping import ClassLookupDict

field_to_converter: ClassLookupDict = ClassLookupDict({})


def converter(converter_class):
    """Decorator to register a converter class"""
    if isinstance(converter_class.field_class, list):
        field_classes = converter_class.field_class
    else:
        field_classes = [converter_class.field_class]
    for field_class in field_classes:
        field_to_converter[field_class] = converter_class()
    return converter_class


def field_to_jsonschema(field):
    if isinstance(field, serializers.Serializer):
        result = to_jsonschema(field)
    else:
        converter = field_to_converter[field]
        result = converter.convert(field)
    if field.label:
        result["title"] = field.label
    if field.help_text:
        result["description"] = field.help_text
    return result


def to_jsonschema(serializer):
    properties = {}
    required = []
    for name, field in serializer.fields.items():
        if field.read_only:
            continue
        sub_schema = field_to_jsonschema(field)
        if field.required:
            required.append(name)
        properties[name] = sub_schema

    result = {"type": "object", "properties": properties}
    if required:
        result["required"] = required
    return result



import pytest
from jsonschema import validate
from rest_framework import serializers

from drf_jsonschema_serializer import to_jsonschema
from testapp.models import Album, Track


@pytest.mark.django_db
def test_string_related_field():
    album = Album.objects.create(
        album_name="Collected Stories", artist="H.P. Lovecraft"
    )
    track1 = track1 = Track.objects.create(
        album=album, order=1, title="The Dunwich Horror", duration=10
    )
    Track.objects.create(album=album, order=2, title="The Call of Cthulhu", duration=15)
    Track.objects.create(
        album=album, order=3, title="At the Mountains of Madness", duration=20
    )

    class AlbumSerializer(serializers.ModelSerializer):
        class Meta:
            model = Album
            fields = ("album_name", "artist", "tracks")

    class TrackSerializer(serializers.ModelSerializer):
        class Meta:
            model = Track
            fields = ("order", "title", "duration")

    json_schema = to_jsonschema(TrackSerializer())

    valid = {"title": "The Dunwich Horror", "order": 1, "duration": 11}

    validate(valid, json_schema)

    serializer = TrackSerializer(track1, data=valid)
    assert serializer.is_valid()

    serializer.save()
    assert track1.duration == 11

    invalid = {"title": "The Dunwich Horror", "order": 1}

    serializer = TrackSerializer(track1, data=invalid)
    assert not serializer.is_valid()

    album_data = AlbumSerializer(album).data

    assert album_data == {
        "album_name": "Collected Stories",
        "artist": "H.P. Lovecraft",
        "tracks": [1, 2, 3],
    }

    album_serializer = AlbumSerializer(album, album_data)
    json_schema = to_jsonschema(album_serializer)

    validate(album_data, json_schema)



from typing import Optional, Type, TypeVar, Union

from jsonschema import Draft202012Validator, FormatChecker, RefResolver
from jsonschema.exceptions import ValidationError as JSONSchemaValidationError
from rest_framework import serializers

from .convert import to_jsonschema

T = TypeVar("T")

_DEFAULT = object()


class JSONSchemaField(serializers.Field):
    """A field that validates incoming data against a JSON Schema.

    This field takes incoming data as Python data structures (objects,
    lists, strings, numbers, bool, None) that is the equivalent of JSON.
    After validation the same structure is returned.

    The data can then be stored in a native JSON field (for instance a
    PostgreSQL ``django.contrib.postgres.JSONField``).
    """

    def __init__(
        self,
        schema: dict,
        resolver: Optional[RefResolver] = None,
        format_checker: Union[FormatChecker, object] = _DEFAULT,
        *args,
        **kwargs,
    ):
        super(JSONSchemaField, self).__init__(*args, **kwargs)
        Draft202012Validator.check_schema(schema)
        self.schema = schema
        if format_checker is _DEFAULT:
            format_checker = FormatChecker()
        self.validator = Draft202012Validator(schema, resolver, format_checker)

    def to_representation(self, obj: T) -> T:
        return obj

    def to_internal_value(self, data: T) -> T:
        try:
            self.validator.validate(data)
        except JSONSchemaValidationError as e:
            raise serializers.ValidationError(e.message)

        return data


class SerializerJSONField(serializers.Field):
    """A field that stores JSON but uses a serializer to validate."""

    def __init__(self, serializer_class: Type[serializers.Serializer], *args, **kw):
        super(SerializerJSONField, self).__init__(*args, **kw)
        self.serializer_class = serializer_class
        self.schema = to_jsonschema(serializer_class())

    def to_representation(self, obj: T) -> T:
        return obj

    def to_internal_value(self, data):
        serializer = self.serializer_class(data=data)
        if not serializer.is_valid():
            serializer.run_validation(data)
        return serializer.data



# import the converters to register them
from . import converters
from .convert import to_jsonschema
from .fields import JSONSchemaField, SerializerJSONField

__all__ = ["JSONSchemaField", "SerializerJSONField", "to_jsonschema", "converters"]



# convert a serializer to a JSON Schema.
from typing import Any, Dict, List, Sequence, Type, Union

from rest_framework import fields, serializers
from rest_framework.settings import api_settings

from .convert import converter, field_to_jsonschema
from .fields import JSONSchemaField, SerializerJSONField

FieldCls = Type[fields.Field]
FieldClass = Union[FieldCls, List[FieldCls]]


class Error(Exception):
    pass


class Converter:
    type: str = ""
    field_class: FieldClass

    def convert(self, field) -> Dict[str, Any]:
        assert self.type, "A type must be specified"
        type: Union[str, List[str]]
        if field.allow_null:
            type = [self.type, "null"]
        else:
            type = self.type
        return {"type": type}


class FormatConverter(Converter):
    type = "string"
    format: str = ""

    def convert(self, field):
        result = super(FormatConverter, self).convert(field)
        if self.format:
            result["format"] = self.format
        return result


@converter
class CharFieldConverter(FormatConverter):
    type = "string"
    field_class: FieldClass = serializers.CharField

    def convert(self, field):
        result = super(CharFieldConverter, self).convert(field)
        if field.max_length is not None:
            result["maxLength"] = field.max_length
        min_length = field.min_length
        if not min_length and not field.allow_blank:
            min_length = 1
        if min_length is not None:
            result["minLength"] = min_length
        return result


@converter
class EmailFieldConverter(CharFieldConverter):
    format = "email"
    field_class = serializers.EmailField


@converter
class RegexFieldConverter(CharFieldConverter):
    field_class = [serializers.RegexField, serializers.SlugField]

    def convert(self, field):
        result = super(RegexFieldConverter, self).convert(field)
        # rely on a lot of internal details...
        result["pattern"] = str(field.validators[-1].regex.pattern)
        return result


@converter
class URLFieldConverter(CharFieldConverter):
    format = "uri"
    field_class = serializers.URLField


@converter
class BooleanFieldConverter(Converter):
    type = "boolean"
    field_class = [serializers.BooleanField]


@converter
class FloatFieldConverter(Converter):
    type = "number"
    field_class: FieldClass = serializers.FloatField

    def convert(self, field):
        result = super(FloatFieldConverter, self).convert(field)
        if field.min_value is not None:
            result["minimum"] = field.min_value
        if field.max_value is not None:
            result["maximum"] = field.max_value
        return result


@converter
class IntegerFieldConverter(FloatFieldConverter):
    type = "integer"
    field_class = serializers.IntegerField


@converter
class DecimalFieldConverter(Converter):
    # the JSON Schema spec doesn't support decimals, I suggested it should
    # https://github.com/json-schema-org/json-schema-spec/issues/361
    type = "string"
    field_class = serializers.DecimalField

    def convert(self, field):
        if field.max_digits is not None:
            raise Error("max_digits not yet supported")
        if not getattr(field, "coerce_to_string", True):
            raise Error("coerce_to_string must be True")
        result = super(DecimalFieldConverter, self).convert(field)
        result["pattern"] = "^\\-?[0-9]*(\\.[0-9]{1,%d})?$" % (field.decimal_places)
        return result


class BaseDateTimeFieldConverter(FormatConverter):
    type = "string"
    format = "date-time"

    expected_input_formats: List[str]
    settings_format: str
    settings_input_formats: Sequence[str]

    def convert(self, field):
        # ugh had to copy from DRF
        format = getattr(self, "format", self.settings_format)
        # FIXME: I mysteriously get out 'date-time' instead of 'iso-8601'...
        if format not in self.expected_input_formats:
            raise Error("format not supported")
        # ugh copy from DRF
        input_formats = getattr(self, "input_formats", self.settings_input_formats)
        if "iso-8601" not in input_formats:
            raise Error("formats beside iso-8601 not supported")
        return super(BaseDateTimeFieldConverter, self).convert(field)


@converter
class DateTimeFieldConverter(BaseDateTimeFieldConverter):
    type = "string"
    format = "date-time"
    field_class = serializers.DateTimeField

    expected_input_formats = ["iso-8601", "date-time"]
    settings_format = api_settings.DATETIME_FORMAT
    settings_input_formats = api_settings.DATETIME_INPUT_FORMATS


@converter
class DateFieldConverter(BaseDateTimeFieldConverter):
    type = "string"
    format = "date"
    field_class = serializers.DateField

    expected_input_formats = ["iso-8601", "date"]
    settings_format = api_settings.DATE_FORMAT
    settings_input_formats = api_settings.DATE_INPUT_FORMATS


@converter
class ChoiceField:
    field_class = serializers.ChoiceField

    def convert(self, field):
        types = set()
        enum = []
        enumNames = []
        if field.allow_blank:
            enum.append("")
            enumNames.append("")
        has_display_names = False
        for choice, display_name in field.choices.items():
            if choice != display_name:
                has_display_names = True
            enum.append(choice)
            enumNames.append(display_name)
            # FIXME: what about choices such as datetime?
            # maybe we need a way to turn concrete instances into
            # the underlying DRF fields just for the sake of
            # conversion. But does a datetime choice make sense?
            if isinstance(choice, str):
                types.add("string")
            elif isinstance(choice, bool):
                types.add("boolean")
            elif isinstance(choice, int):
                types.add("integer")
            elif isinstance(choice, float):
                types.add("number")
            elif choice is None:
                types.add("null")
        if field.allow_null:
            types.add("null")
            if enum[0] is not None:
                enum.insert(0, None)
                enumNames.insert(0, "")
            has_display_names = True

        types = sorted(list(types))
        if len(types) == 1:
            type = types[0]
        else:
            type = types
        result = {"type": type, "enum": enum}
        if has_display_names:
            result["enumNames"] = enumNames
        return result


@converter
class ListField:
    field_class = serializers.ListField

    def convert(self, field):
        result = {"type": "array", "items": field_to_jsonschema(field.child)}
        if field.min_length is not None:
            result["minItems"] = field.min_length
        if field.max_length is not None:
            result["maxItems"] = field.max_length
        return result


@converter
class DictField:
    field_class = serializers.DictField

    def convert(self, field):
        return {
            "type": "object",
            "additionalProperties": field_to_jsonschema(field.child),
        }


@converter
class JSONSchemaFieldConverter:
    field_class = JSONSchemaField

    def convert(self, field):
        return field.schema


@converter
class SerializerJSONFieldConverter:
    field_class = SerializerJSONField

    def convert(self, field):
        return field.schema


@converter
class ManyRelatedFieldConverter:
    field_class = serializers.ManyRelatedField

    def convert(self, field):
        return {"type": "array", "items": field_to_jsonschema(field.child_relation)}


@converter
class PrimaryKeyRelatedFieldConverter:
    field_class = serializers.PrimaryKeyRelatedField

    def convert(self, field):
        return {"type": "integer"}


@converter
class StringRelatedFieldConverter:
    field_class = serializers.StringRelatedField

    def convert(self, field):
        return {"type": "string"}


@converter
class HyperlinkedRelatedFieldConverter:
    field_class = serializers.HyperlinkedRelatedField

    def convert(self, field):
        return {"type": "string", "format": "uri"}


@converter
class SlugRelatedFieldConverter:
    field_class = serializers.SlugRelatedField

    def convert(self, field):
        # FIXME: hardcoded slug regex
        return {"type": "string", "pattern": "^[-a-zA-Z0-9_]+$"}


@converter
class ListSerializerConverter:
    field_class = serializers.ListSerializer

    def convert(self, field):
        result = {"type": "array", "items": field_to_jsonschema(field.child)}
        if not field.allow_empty:
            result["minItems"] = 1
        return result



import pytest
from jsonschema.exceptions import ValidationError as JSONSchemaValidationError
from rest_framework import serializers

from drf_jsonschema_serializer import (
    JSONSchemaField,
    SerializerJSONField,
    to_jsonschema,
)
from testapp.models import Album, Track

registered_serializer_classes = []


def register(serializer):
    registered_serializer_classes.append(serializer)
    return serializer


@register
class BooleanFieldBasic(serializers.Serializer):
    foo = serializers.BooleanField()

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "boolean", "title": "Foo"}},
        "required": ["foo"],
    }

    valid = [{"foo": True}, {"foo": False}]
    invalid = [{}, {"foo": "something"}, {"foo": None}]
    # the serializer actually accepts 0 and 1 and "true" and "false"
    # and even "0" and "1"
    invalid_by_schema = [
        {"foo": 0},
        {"foo": 1},
        {"foo": "true"},
        {"foo": "false"},
        {"foo": "0"},
        {"foo": "1"},
    ]


@register
class BooleanFieldNotRequired(serializers.Serializer):
    foo = serializers.BooleanField(required=False)

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "boolean", "title": "Foo"}},
    }

    valid = [{"foo": True}, {"foo": False}, {}]
    invalid = [{"foo": "something"}, {"foo": None}]
    invalid_by_schema = [{"foo": 0}, {"foo": 1}]


@register
class NullBooleanFieldBasic(serializers.Serializer):
    foo = serializers.BooleanField(allow_null=True)

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": ["boolean", "null"], "title": "Foo"}},
        "required": ["foo"],
    }

    valid = [{"foo": True}, {"foo": False}, {"foo": None}]
    invalid = [{}, {"foo": "something"}, {"foo": 10}]


@register
class CharFieldBasic(serializers.Serializer):
    foo = serializers.CharField()

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "string", "minLength": 1, "title": "Foo"}},
        "required": ["foo"],
    }

    valid = [{"foo": "something"}]
    invalid = [{}, {"foo": ""}, {"foo": None}, {"foo": False}]
    invalid_by_schema = [{"foo": 1}]


@register
class CharFieldMaxLength(serializers.Serializer):
    foo = serializers.CharField(max_length=10)

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": "string", "minLength": 1, "maxLength": 10, "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": "x" * 10}]
    invalid = [{}, {"foo": ""}, {"foo": "x" * 11}, {"foo": None}]


@register
class CharFieldAllowNull(serializers.Serializer):
    foo = serializers.CharField(allow_null=True)

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": ["string", "null"], "minLength": 1, "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": "something"}, {"foo": None}]
    invalid = [{}, {"foo": ""}]


@register
class CharFieldNotRequired(serializers.Serializer):
    foo = serializers.CharField(required=False)

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "string", "minLength": 1, "title": "Foo"}},
    }

    valid = [{"foo": "something"}, {}]
    invalid = [{"foo": None}, {"foo": ""}]


@register
class EmailFieldBasic(serializers.Serializer):
    foo = serializers.EmailField()

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": "string", "minLength": 1, "format": "email", "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": "someone@example.com"}]
    invalid = [{}, {"foo": ""}, {"foo": None}, {"foo": "someone"}]


@register
class RegexFieldBasic(serializers.Serializer):
    foo = serializers.RegexField(regex="^[0-9]+$")

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "string",
                "minLength": 1,
                "pattern": "^[0-9]+$",
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": "1"}, {"foo": "123"}]
    invalid = [{}, {"foo": ""}, {"foo": None}, {"foo": "someone"}, {"foo": "12foo"}]


@register
class SlugFieldBasic(serializers.Serializer):
    foo = serializers.SlugField()

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "string",
                "minLength": 1,
                "pattern": "^[-a-zA-Z0-9_]+$",
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": "1"}, {"foo": "123"}, {"foo": "foo12"}, {"foo": "12something"}]
    invalid = [{}, {"foo": ""}, {"foo": None}, {"foo": "$$$"}, {"foo": "!foo"}]


@register
class URLFieldBasic(serializers.Serializer):
    foo = serializers.URLField()

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": "string", "minLength": 1, "format": "uri", "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": "http://example.com"}]
    invalid = [{"foo": "something"}, {"foo": "/foo/bar"}, {"foo": ""}]


@register
class IntegerField(serializers.Serializer):
    foo = serializers.IntegerField()

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "integer", "title": "Foo"}},
        "required": ["foo"],
    }

    valid = [{"foo": 1}]
    invalid = [{}, {"foo": "something"}, {"foo": None}, {"foo": 1.5}, {"foo": "10.5"}]
    # the serializer actually accepts strings
    invalid_by_schema = [{"foo": "10"}]


@register
class IntegerFieldMinValue(serializers.Serializer):
    foo = serializers.IntegerField(min_value=10)

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "integer", "minimum": 10, "title": "Foo"}},
        "required": ["foo"],
    }

    valid = [{"foo": 10}, {"foo": 11}]
    invalid = [
        {},
        {"foo": "something"},
        {"foo": None},
        {"foo": 1},
        {"foo": 1.5},
        {"foo": 9},
        {"foo": 10.5},
    ]


@register
class IntegerFieldMaxValue(serializers.Serializer):
    foo = serializers.IntegerField(max_value=10)

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "integer", "maximum": 10, "title": "Foo"}},
        "required": ["foo"],
    }

    valid = [{"foo": 10}, {"foo": 1}]
    invalid = [
        {},
        {"foo": "something"},
        {"foo": None},
        {"foo": 11},
        {"foo": 1.5},
        {"foo": 10.5},
    ]


@register
class FloatField(serializers.Serializer):
    foo = serializers.FloatField()

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "number", "title": "Foo"}},
        "required": ["foo"],
    }

    valid = [{"foo": 10}, {"foo": 10.5}]
    invalid = [{}, {"foo": "something"}, {"foo": None}]
    # the serializer actually accepts strings
    invalid_by_schema = [{"foo": "1.5"}]


@register
class FloatFieldMinValue(serializers.Serializer):
    foo = serializers.FloatField(min_value=10)

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "number", "minimum": 10, "title": "Foo"}},
        "required": ["foo"],
    }

    valid = [{"foo": 10}, {"foo": 10.5}]
    invalid = [
        {},
        {"foo": "something"},
        {"foo": None},
        {"foo": 1},
        {"foo": 1.5},
        {"foo": 9.9},
    ]


@register
class FloatFieldMaxValue(serializers.Serializer):
    foo = serializers.FloatField(max_value=10)

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "number", "maximum": 10, "title": "Foo"}},
        "required": ["foo"],
    }

    valid = [{"foo": 9}, {"foo": 9.5}]
    invalid = [{}, {"foo": "something"}, {"foo": None}, {"foo": 10.5}, {"foo": 11}]


@register
class DecimalField(serializers.Serializer):
    # note that max_digits is explicitly not supported
    # and that we don't support min_value & max_value either due
    # to limitations in JSON-Schema
    foo = serializers.DecimalField(max_digits=None, decimal_places=2)

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "string",
                "pattern": "^\\-?[0-9]*(\\.[0-9]{1,2})?$",
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [
        {"foo": "10"},
        {"foo": "10.5"},
        {"foo": "-1.5"},
        {"foo": "9999999999.5"},
        {"foo": ".5"},
    ]
    invalid = [{}, {"foo": "something"}, {"foo": None}, {"foo": "1.123"}]
    # the serializer actually accepts numbers too
    # (but a string is used by the JSON if coerce_to_string is True,
    # the default)
    invalid_by_schema = [{"foo": 1}, {"foo": 1.5}]


@register
class DateTimeField(serializers.Serializer):
    foo = serializers.DateTimeField()

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": "string", "format": "date-time", "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": "2017-01-12T16:01:56Z"}]
    invalid = [{}, {"foo": "something"}, {"foo": "2017-14-1T16:01:56Z"}]
    # JSON schema is strict about date-time format, but the serializer
    # accepts shorter formats
    invalid_by_schema = [{"foo": "2017-01-12T16:01"}]


@register
class DateField(serializers.Serializer):
    foo = serializers.DateField()

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "string", "format": "date", "title": "Foo"}},
        "required": ["foo"],
    }

    valid = [{"foo": "2017-01-12"}]
    invalid = [
        {},
        {"foo": "something"},
        {"foo": "2017-01-12T16:01:00:Z"},
        {"foo": "2017-01"},
        {"foo": "2017-14-01"},
    ]


@register
class ChoiceFieldStrings(serializers.Serializer):
    foo = serializers.ChoiceField(["a", "b", "c"])

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": "string", "enum": ["a", "b", "c"], "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": "a"}, {"foo": "b"}, {"foo": "c"}]
    invalid = [{}, {"foo": None}, {"foo": ""}, {"foo": "d"}, {"foo": 1}]


@register
class ChoiceFieldStringsAllowBlank(serializers.Serializer):
    foo = serializers.ChoiceField(["a", "b", "c"], allow_blank=True)

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": "string", "enum": ["", "a", "b", "c"], "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": "a"}, {"foo": "b"}, {"foo": "c"}, {"foo": ""}]
    invalid = [{}, {"foo": None}, {"foo": "d"}, {"foo": 1}]


@register
class ChoiceFieldStringsBlankChoice(serializers.Serializer):
    foo = serializers.ChoiceField(["", "a", "b", "c"])

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": "string", "enum": ["", "a", "b", "c"], "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": "a"}, {"foo": "b"}, {"foo": "c"}, {"foo": ""}]
    invalid = [{}, {"foo": None}, {"foo": "d"}, {"foo": 1}]


@register
class ChoiceFieldStringsAllowNull(serializers.Serializer):
    foo = serializers.ChoiceField(["a", "b", "c"], allow_null=True)

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": ["null", "string"],
                "enum": [None, "a", "b", "c"],
                "enumNames": ["", "a", "b", "c"],
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": "a"}, {"foo": "b"}, {"foo": "c"}, {"foo": None}]
    invalid = [{}, {"foo": ""}, {"foo": "d"}, {"foo": 1}]


@register
class ChoiceFieldStringsNullChoice(serializers.Serializer):
    foo = serializers.ChoiceField([(None, "Nothing"), "a", "b", "c"], allow_null=True)

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": ["null", "string"],
                "enum": [None, "a", "b", "c"],
                "enumNames": ["Nothing", "a", "b", "c"],
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": "a"}, {"foo": "b"}, {"foo": "c"}, {"foo": None}]
    invalid = [{}, {"foo": ""}, {"foo": "d"}, {"foo": 1}]


@register
class ChoiceFieldDisplayNames(serializers.Serializer):
    foo = serializers.ChoiceField([("a", "A"), ("b", "B"), ("c", "C")])

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "string",
                "enum": ["a", "b", "c"],
                "enumNames": ["A", "B", "C"],
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": "a"}, {"foo": "b"}, {"foo": "c"}]
    invalid = [{}, {"foo": None}, {"foo": ""}, {"foo": "d"}, {"foo": 1}]


@register
class ChoiceFieldInts(serializers.Serializer):
    foo = serializers.ChoiceField([1, 2, 3])

    json_schema = {
        "type": "object",
        "properties": {"foo": {"type": "integer", "enum": [1, 2, 3], "title": "Foo"}},
        "required": ["foo"],
    }

    valid = [{"foo": 1}, {"foo": 2}, {"foo": 3}]
    invalid = [{}, {"foo": None}, {"foo": ""}, {"foo": "d"}, {"foo": 4}]


@register
class ChoiceFieldIntsDisplaynames(serializers.Serializer):
    foo = serializers.ChoiceField([(1, "One"), (2, "Two"), (3, "Three")])

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "integer",
                "enum": [1, 2, 3],
                "enumNames": ["One", "Two", "Three"],
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": 1}, {"foo": 2}, {"foo": 3}]
    invalid = [{}, {"foo": None}, {"foo": ""}, {"foo": "d"}, {"foo": 4}]


@register
class ChoiceFieldFloats(serializers.Serializer):
    foo = serializers.ChoiceField([1.1, 2.1, 3.1])

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": "number", "enum": [1.1, 2.1, 3.1], "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": 1.1}, {"foo": 2.1}, {"foo": 3.1}]
    invalid = [{}, {"foo": None}, {"foo": ""}, {"foo": "d"}, {"foo": 4}, {"foo": 1.2}]


@register
class ChoiceFieldMixed(serializers.Serializer):
    foo = serializers.ChoiceField(["a", 1])

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": ["integer", "string"], "enum": ["a", 1], "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": 1}, {"foo": "a"}]
    invalid = [{}, {"foo": None}, {"foo": "b"}, {"foo": 3}]


@register
class ChoiceFieldBool(serializers.Serializer):
    foo = serializers.ChoiceField([False, True])

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": "boolean", "enum": [False, True], "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": False}, {"foo": True}]
    invalid = [{}, {"foo": None}, {"foo": ""}, {"foo": "d"}, {"foo": 1}]


class SubSerializer(serializers.Serializer):
    foo = serializers.IntegerField()
    bar = serializers.CharField()


@register
class SubSerializerBasic(serializers.Serializer):
    sub = SubSerializer(label="Sub serializer!")
    another = serializers.CharField()

    json_schema = {
        "type": "object",
        "properties": {
            "sub": {
                "type": "object",
                "title": "Sub serializer!",
                "properties": {
                    "foo": {"type": "integer", "title": "Foo"},
                    "bar": {"type": "string", "minLength": 1, "title": "Bar"},
                },
                "required": ["foo", "bar"],
            },
            "another": {"type": "string", "minLength": 1, "title": "Another"},
        },
        "required": ["sub", "another"],
    }

    valid = [{"sub": {"foo": 3, "bar": "BAR"}, "another": "HALLO"}]
    invalid = [
        {},
        {"sub": "what", "another": "HALLO"},
        {"sub": {"foo": 3, "bar": ""}, "another": "HALLO"},
    ]


@register
class SubSerializerMany(serializers.Serializer):
    sub = SubSerializer(many=True)
    another = serializers.CharField()

    json_schema = {
        "type": "object",
        "properties": {
            "sub": {
                "type": "array",
                "title": "Sub",
                "items": {
                    "type": "object",
                    "properties": {
                        "foo": {"type": "integer", "title": "Foo"},
                        "bar": {"type": "string", "minLength": 1, "title": "Bar"},
                    },
                    "required": ["foo", "bar"],
                },
            },
            "another": {"type": "string", "minLength": 1, "title": "Another"},
        },
        "required": ["sub", "another"],
    }

    valid = [
        {"sub": [{"foo": 3, "bar": "BAR"}], "another": "HALLO"},
        {"sub": [], "another": "HALLO"},
    ]
    invalid = [
        {},
        {"sub": "what", "another": "HALLO"},
        {"sub": [{"foo": 3, "bar": ""}], "another": "HALLO"},
    ]


@register
class SubSerializerManyNotAllowEmpty(serializers.Serializer):
    sub = SubSerializer(many=True, allow_empty=False)
    another = serializers.CharField()

    json_schema = {
        "type": "object",
        "properties": {
            "sub": {
                "type": "array",
                "title": "Sub",
                "minItems": 1,
                "items": {
                    "type": "object",
                    "properties": {
                        "foo": {"type": "integer", "title": "Foo"},
                        "bar": {"type": "string", "minLength": 1, "title": "Bar"},
                    },
                    "required": ["foo", "bar"],
                },
            },
            "another": {"type": "string", "minLength": 1, "title": "Another"},
        },
        "required": ["sub", "another"],
    }

    valid = [{"sub": [{"foo": 3, "bar": "BAR"}], "another": "HALLO"}]
    invalid = [
        {},
        {"sub": [], "another": "HALLO"},
        {"sub": "what", "another": "HALLO"},
        {"sub": [{"foo": 3, "bar": ""}], "another": "HALLO"},
    ]


@register
class ListFieldBasic(serializers.Serializer):
    foo = serializers.ListField(child=serializers.IntegerField())

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": "array", "items": {"type": "integer"}, "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": [1, 2, 3]}, {"foo": []}]
    invalid = [{}, {"foo": None}, {"foo": 1}, {"foo": ""}]


class IntegerListField(serializers.ListField):
    child = serializers.IntegerField()


@register
class ListFieldDeclarative(serializers.Serializer):
    foo = IntegerListField()

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {"type": "array", "items": {"type": "integer"}, "title": "Foo"}
        },
        "required": ["foo"],
    }

    valid = [{"foo": [1, 2, 3]}, {"foo": []}]
    invalid = [{}, {"foo": None}, {"foo": 1}, {"foo": ""}]


@register
class ListFieldMinLength(serializers.Serializer):
    foo = serializers.ListField(child=serializers.IntegerField(), min_length=1)

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "array",
                "items": {"type": "integer"},
                "minItems": 1,
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": [1, 2, 3]}, {"foo": [1]}]
    invalid = [{}, {"foo": None}, {"foo": 1}, {"foo": ""}, {"foo": []}]


@register
class ListFieldMaxLength(serializers.Serializer):
    foo = serializers.ListField(child=serializers.IntegerField(), max_length=2)

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "array",
                "items": {"type": "integer"},
                "maxItems": 2,
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": [1, 2]}, {"foo": []}]
    invalid = [{}, {"foo": None}, {"foo": 1}, {"foo": ""}, {"foo": [1, 2, 3]}]


@register
class ListFieldObject(serializers.Serializer):
    foo = serializers.ListField(child=SubSerializer())

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "array",
                "title": "Foo",
                "items": {
                    "type": "object",
                    "properties": {
                        "foo": {"type": "integer", "title": "Foo"},
                        "bar": {"type": "string", "minLength": 1, "title": "Bar"},
                    },
                    "required": ["foo", "bar"],
                },
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": [{"foo": 1, "bar": "BAR"}, {"foo": 2, "bar": "BAR2"}]}]
    invalid = [
        {},
        {"foo": None},
        {"foo": [{"foo": 1, "bar": False}]},
        {"foo": [{"foo": 1}]},
    ]


@register
class DictFieldBasic(serializers.Serializer):
    foo = serializers.DictField(child=serializers.IntegerField())

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "object",
                "additionalProperties": {"type": "integer"},
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": {"a": 1, "b": 2}}, {"foo": {}}]
    invalid = [
        {},
        {"foo": None},
        {"foo": 1},
        {"foo": {"a": "nointeger"}},
    ]


@register
class DictFieldObject(serializers.Serializer):
    foo = serializers.DictField(child=SubSerializer())

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "object",
                "additionalProperties": {
                    "type": "object",
                    "properties": {
                        "foo": {"type": "integer", "title": "Foo"},
                        "bar": {"type": "string", "minLength": 1, "title": "Bar"},
                    },
                    "required": ["foo", "bar"],
                },
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": {"a": {"foo": 1, "bar": "BAR"}, "b": {"foo": 2, "bar": "BAR2"}}}]
    invalid = [
        {},
        {"foo": None},
        {"foo": 1},
        {"foo": {"a": {"foo": 1, "bar": False}}},
    ]


@register
class JSONSchemaFieldBasic(serializers.Serializer):
    foo = JSONSchemaField(
        schema={"type": "object", "properties": {"a": {"type": "integer"}}}
    )

    # FIXME: is it correct to add the title here?
    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "object",
                "properties": {"a": {"type": "integer"}},
                "title": "Foo",
            }
        },
        "required": ["foo"],
    }

    valid = [{"foo": {"a": 1}}, {"foo": {}}]
    invalid = [
        {},
        {"foo": None},
        {"foo": 1},
        {"foo": {"a": "nointeger"}},
    ]


@register
class ModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = Album
        fields = ("album_name", "artist")

    json_schema = {
        "type": "object",
        "properties": {
            "album_name": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Album Name",  # note we take title from model field
            },
            "artist": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Artist",
            },
        },
        "required": ["album_name", "artist"],
    }

    valid = []
    invalid = []


@register
class ModelSerializerWithRelated(serializers.ModelSerializer):
    class Meta:
        model = Album
        fields = ("album_name", "artist", "tracks")

    json_schema = {
        "type": "object",
        "properties": {
            "album_name": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Album Name",
            },
            "artist": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Artist",
            },
            "tracks": {
                "type": "array",
                "title": "Tracks",
                "items": {"type": "integer"},
            },
        },
        "required": ["album_name", "artist", "tracks"],
    }

    valid = []
    invalid = []


class TrackSerializer(serializers.ModelSerializer):
    class Meta:
        model = Track
        fields = ("order", "title", "duration")


@register
class ModelSerializerWithRelatedSerializer(serializers.ModelSerializer):
    tracks = TrackSerializer(many=True)

    class Meta:
        model = Album
        fields = ("album_name", "artist", "tracks")

    json_schema = {
        "type": "object",
        "properties": {
            "album_name": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Album Name",
            },
            "artist": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Artist",
            },
            "tracks": {
                "type": "array",
                "title": "Tracks",
                "items": {
                    "type": "object",
                    "properties": {
                        "order": {
                            "type": "integer",
                            "title": "Order",
                            "description": "The order of the track",
                        },
                        "title": {
                            "type": "string",
                            "minLength": 1,
                            "maxLength": 100,
                            "title": "Title",
                        },
                        "duration": {"type": "integer", "title": "Duration"},
                    },
                    "required": ["order", "title", "duration"],
                },
            },
        },
        "required": ["album_name", "artist", "tracks"],
    }

    valid = []
    invalid = []


@register
class PrimaryKeyRelatedField(serializers.ModelSerializer):
    # album is a foreign key
    class Meta:
        model = Track
        fields = ("album", "order", "title", "duration")

    json_schema = {
        "type": "object",
        "properties": {
            "album": {"type": "integer", "title": "Album"},
            "order": {
                "type": "integer",
                "title": "Order",
                "description": "The order of the track",
            },
            "title": {
                "type": "string",
                "title": "Title",
                "maxLength": 100,
                "minLength": 1,
            },
            "duration": {"type": "integer", "title": "Duration"},
        },
        "required": ["album", "order", "title", "duration"],
    }

    valid = []
    invalid = []


@register
class StringRelatedField(serializers.ModelSerializer):
    tracks = serializers.StringRelatedField(many=True)

    class Meta:
        model = Album
        fields = ("album_name", "artist", "tracks")

    json_schema = {
        "type": "object",
        "properties": {
            "album_name": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Album Name",
            },
            "artist": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Artist",
            },
            "tracks": {"type": "array", "title": "Tracks", "items": {"type": "string"}},
        },
        "required": ["album_name", "artist", "tracks"],
    }

    valid = []
    invalid = []


@register
class HyperlinkedRelatedField(serializers.ModelSerializer):
    tracks = serializers.HyperlinkedRelatedField(
        many=True, queryset=Track.objects, view_name="fake"
    )

    class Meta:
        model = Album
        fields = ("album_name", "artist", "tracks")

    json_schema = {
        "type": "object",
        "properties": {
            "album_name": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Album Name",
            },
            "artist": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Artist",
            },
            "tracks": {
                "type": "array",
                "title": "Tracks",
                "items": {"type": "string", "format": "uri"},
            },
        },
        "required": ["album_name", "artist", "tracks"],
    }

    valid = []
    invalid = []


@register
class SlugRelatedField(serializers.ModelSerializer):
    tracks = serializers.SlugRelatedField(
        many=True, queryset=Track.objects, slug_field="title"
    )

    class Meta:
        model = Album
        fields = ("album_name", "artist", "tracks")

    json_schema = {
        "type": "object",
        "properties": {
            "album_name": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Album Name",
            },
            "artist": {
                "type": "string",
                "maxLength": 100,
                "minLength": 1,
                "title": "Artist",
            },
            "tracks": {
                "type": "array",
                "title": "Tracks",
                "items": {"type": "string", "pattern": "^[-a-zA-Z0-9_]+$"},
            },
        },
        "required": ["album_name", "artist", "tracks"],
    }

    valid = []
    invalid = []


@register
class ReadOnlyIsNotConverted(serializers.Serializer):
    foo = serializers.CharField(read_only=True)

    json_schema = {"type": "object", "properties": {}}

    valid = [{"foo": "something"}, {}, {"foo": ""}, {"foo": None}, {"foo": False}]
    invalid = []


@register
class HelpTextIsDescription(serializers.Serializer):
    foo = serializers.CharField(help_text="This is really a foo")

    json_schema = {
        "type": "object",
        "properties": {
            "foo": {
                "type": "string",
                "minLength": 1,
                "title": "Foo",
                "description": "This is really a foo",
            }
        },
        "required": ["foo"],
    }

    valid = []
    invalid = []


class SJFSerializer(serializers.Serializer):
    foo = serializers.IntegerField()


@register
class SerializerJSON(serializers.Serializer):
    json = SerializerJSONField(SJFSerializer)

    json_schema = {
        "type": "object",
        "properties": {
            "json": {
                "type": "object",
                "title": "Json",
                "properties": {"foo": {"type": "integer", "title": "Foo"}},
                "required": ["foo"],
            }
        },
        "required": ["json"],
    }

    valid = [{"json": {"foo": 1}}]
    invalid = [{"json": {"foo": "Not a number"}}]
    invalid_by_schema = [{"json": {"foo": "1"}}]


def test_correct_generated_schema(serializer_class, expected_json_schema):
    generated_json_schema = to_jsonschema(serializer_class())
    assert generated_json_schema == expected_json_schema


def test_valid_by_serializer(serializer_class, valid):
    serializer_class(data=valid).run_validation(valid)


def test_invalid_by_serializer(serializer_class, invalid):
    with pytest.raises(serializers.ValidationError):
        serializer_class(data=invalid).run_validation(invalid)


def test_valid_by_jsonschema(validator, valid):
    validator.validate(valid)


def test_invalid_by_jsonschema(validator, invalid):
    with pytest.raises(JSONSchemaValidationError):
        validator.validate(invalid)
